from qugrad.pulses import compose, compose_unpack

def test_compose():
    def g(x):
        return x + 3

    def f(x):
        return x ** 2

    h = compose(f, g)
    assert h(3) != g(f(3)) # check the test functions don't commute
    assert h(3) == f(g(3)) # check the composition is correct

def test_compose_additional_args():
    def g(x):
        return x + 3

    def f(x, a, power):
        return x ** power + a

    h = compose(f, g, 5, power=3)
    assert h(3) == f(g(3), 5, power=3)

def test_compose_uppack():
    def g(x, y):
        return x + y, x*y

    def f(x, y):
        return x ** y

    h = compose_unpack(f, g)
    try:
        assert h(2, 3) != g(*f(2, 3)) # check the test functions don't commute
    except TypeError:
        pass
    assert h(2, 3) == f(*g(2, 3)) # check the composition is correct

def test_compose_uppack_additional_args():
    def g(x, y):
        return x + y, x*y

    def f(x, y, a, b):
        return x ** y + a*b

    h = compose_unpack(f, g, 5, b=3)
    assert h(2, 3) == f(*g(2, 3), 5, b=3)


import os
import toml
import yaml

PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
README_FILE = os.path.join(PROJECT_DIR, "README.md")
TOML_FILE = os.path.join(PROJECT_DIR, "pyproject.toml")
CITATION_FILE = os.path.join(PROJECT_DIR, "CITATION.cff")
CHANGELOG_FILE = os.path.join(PROJECT_DIR, "ChangeLog.md")

def test_version_consistency():
    toml_version = toml.load(TOML_FILE)["project"]["version"]
    with open(CITATION_FILE, "r") as f:
        citation_content = yaml.safe_load(f)
    citation_version = citation_content["version"]
    assert citation_version == toml_version
    with open(README_FILE, "r") as f:
        for line in f.readline():
            if "`](ChangeLog.md#release-" in line:
                assert f"[`{toml_version}`](ChangeLog.md#release-{toml_version.replace('.', '')})" in line
    with open(CHANGELOG_FILE, "r") as f:
        for line in f.readlines():
            if line.startswith("## "):
                assert toml_version in line
                break
        else: assert False, "Version not found in CHANGELOG.md"


import numpy as np

from qugrad import QuantumSystem, HilbertSpace
from qugrad.systems import TransformedSystem

from .pauli_matrices import X, Y, Z

def test_TransformedSystem_initialisation():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    assert quantum_system.original_system == original_system
    assert np.array_equal(quantum_system.H0, H02)
    assert np.array_equal(quantum_system.Hs, Hs2)
    assert quantum_system.hilbert_space == hilbert_space2
    assert quantum_system.base_system == original_system
    

def test_original_system_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    original_system2 = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.original_system = original_system2
    except AttributeError:
        pass
    else:
        raise AssertionError("TransformedSystem.original_system should be read-only")

def test_base_system_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    original_system2 = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.base_system = original_system2
    except AttributeError:
        pass
    else:
        raise AssertionError("TransformedSystem.base_system should be read-only")

def test_base_system():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    H03 = X
    Hs3 = [Y, Z]
    hilbert_space3 = HilbertSpace([1, 2])
    quantum_system2 = TransformedSystem(quantum_system,
                                        H03,
                                        Hs3,
                                        hilbert_space3)
    assert quantum_system2.base_system == original_system
    assert quantum_system2.original_system == quantum_system
    

def test_pre_processing():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    output1 = quantum_system._pre_processing(*args)
    output2 = original_system._pre_processing(*args)
    assert len(output1) == len(output2)
    for out1, out2 in zip(output1, output2):
        assert np.array_equal(out1, out2)
    

def test_envolope_processing():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    H02 = X
    Hs2 = [Y, Z]
    hilbert_space2 = HilbertSpace([1, 2])
    quantum_system = TransformedSystem(original_system,
                                       H02,
                                       Hs2,
                                       hilbert_space2)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, dt, frequencies, number_channels)
    out1 = quantum_system._envolope_processing(*args)
    out2 = original_system._envolope_processing(*args)
    assert np.array_equal(out1, out2)


"""
Defines a class representing a Hilbert space.
"""

import numpy as np
from typing import Optional, Any, Union

class HilbertSpace():
    """
    Represents a Hilbert space with a prefered basis indexed by non-negative
    integers.

    Note
    ----
    The class also can be treated as an iterator and will return :attr:`basis`
    as the iterator.


    ---
    """
    
    _basis: np.ndarray[int]
    "An array of positive integers labeling the Hilbert space basis"
    _inverse: np.ndarray[int]
    "An array satisfying the property ``_inverse[basis[i]]=i``"
    
    def __init__(self, basis: np.ndarray[int]):
        """
        Initialises a new :class:`HilbertSpace`.

        Parameters
        ----------
        basis : NDArray[Shape[:attr:`dim`], int]
            The labels for the basis vectors.
        """
        self._basis = np.array(basis)
        self._basis.flags.writeable = False
        if len(self._basis) == 0:
            self._inverse = np.empty(0, dtype=np.int64)
        else:
            self._inverse = np.zeros(self._basis.max()+1, dtype=np.int64)
        for i, x in enumerate(self._basis):
            self._inverse[x] = i
        self._inverse.flags.writeable = False
    def __iter__(self):         return iter(self._basis)
    def __getitem__(self, key): return self._basis[key]
    def __len__(self):          return len(self._basis)
    def __eq__(self, other: "HilbertSpace") -> bool:
        if isinstance(other, self.__class__):
            return np.array_equal(self._basis, other._basis)
        return False
    @property
    def dim(self):
        """
        The Hilbert space dimension
        """
        return len(self)
    @property
    def basis(self) -> np.ndarray[int]:
        """
        An array of positive integers labeling the Hilbert space basis 
        """
        return self._basis
    @property
    def inverse(self) -> np.ndarray[int]:
        "An array satisfying the property ``self.inverse[self.basis[i]]=i``"
        return self._inverse
    def basis_vector(self, basis_state_label: int) -> np.ndarray[np.complex128]:
        """Returns a column vector represnetation corresponding to the input
        basis state label.

        Parameters
        ----------
        basis_state_label : int
            A positive integer denoting the label of the basis state to generate 
            the basis vector for.

        Returns
        -------
        NDArray[Shape[:attr:`dim`], np.complex128]
            The basis vector corresponding to the input basis state label.
        """
        return np.eye(1,
                      len(self._basis),
                      self._inverse[basis_state_label]
                     ).flatten().astype(np.complex128)
    def get_subspace(self, filter: np.ndarray[bool]) -> "HilbertSpace":
        """Generates a new subspace by filtering the basis state labels.

        Parameters
        ----------
        filter : NDArray[Shape[:attr:`dim`], bool]
            ``True`` entries are retained in the new subspace.

        Returns
        -------
        HilbertSpace
            The filtered subspace.
        """
        return HilbertSpace(self._basis[filter])
    @staticmethod
    def _labels(state: Any) -> str:
        """
        Generates a strings that represent the state.

        Parameters
        ----------
        state : Any
            The state to label.

        Returns
        -------
        str
            The label for the specified state.
        """
        return f"|{repr(state)}⟩"
    def labels(self,
               states: Optional[Union[int, list[int]]] = None
              ) -> Union[str, list[str]]:
        """
        Generates a string (list of strings) that represent the state(s).

        Parameters
        ----------
        states : int | list[int], optional
            The state(s) to label. If ``None`` then the labels for all states in
            :attr:`basis` are returned. By default ``None``.

        Returns
        -------
        str | list[str]
            The label(s) for the specified states.
        """
        if isinstance(states, int):
            return self._labels(states)
        elif states is None: states = self.basis
        return [self._labels(int(s)) for s in states]


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'QuGrad'
copyright = '2025, Christopher K. Long'
author = 'Christopher K. Long'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.napoleon',
              'sphinx.ext.autosummary',
              'sphinx.ext.viewcode',
              'sphinx.ext.duration',
              'sphinx.ext.mathjax',
              'sphinx_math_dollar',
              'myst_parser',
              'stubdoc',
              'sphinx_tabs.tabs',
              'sphinx_codeautolink',
              'sphinx.ext.intersphinx']
myst_enable_extensions = ["dollarmath"]

autosummary_generate = True # Turn on sphinx.ext.autosummary
autosummary_imported_members = True
# autodoc_inherit_docstrings = True
autodoc_member_order = "groupwise"
napoleon_include_init_with_doc = True
napoleon_numpy_docstring = True
viewcode_line_numbers = True
myst_heading_anchors = 5
suppress_warnings = ["myst.header"]
module_names = ["qugrad.systems._systems"]
sphinx_tabs_disable_tab_closing = True
toc_object_entries_show_parents = 'hide'

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

mathjax3_config = {
    'loader': {'load': ['[tex]/mathtools', '[tex]/physics']},
    'tex': {
        'packages': {'[+]': ['mathtools', 'physics']},
        'inlineMath': [['\\(', '\\)']],
        'displayMath': [['\\[', '\\]']],
    }
}

intersphinx_mapping = {
    "py_ste": (
        "https://PySTE.readthedocs.io/en/latest",
        None,
    ),
}



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']

import os
import sys
import subprocess
sys.path.insert(0, os.path.abspath(os.path.join('..', 'src')))


DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
subprocess.check_call(f"cd {DIR}; cffconvert -f apalike > docs/citation/citation_files/citation.txt", shell=True)
subprocess.check_call(f"cd {DIR}; cffconvert -f bibtex > docs/citation/citation_files/citation.bib", shell=True)
subprocess.check_call(f"cd {DIR}; cffconvert -f ris > docs/citation/citation_files/citation.ris", shell=True)
subprocess.check_call(f"cd {DIR}; cffconvert -f codemeta > docs/citation/citation_files/citation_codemeta.json", shell=True)
subprocess.check_call(f"cd {DIR}; cffconvert -f endnote > docs/citation/citation_files/citation.enw", shell=True)


import numpy as np

from qugrad import QuantumSystem, HilbertSpace
from qugrad.systems import PulseForm

from .pauli_matrices import X, Y, Z

def pulse_function():
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    return ctrl_amp_env, initial_state, dt, frequencies, number_channels

def pulse_function_append(ctrl_amp_env,
                          initial_state,
                          dt,
                          frequencies,
                          number_channels):
    return 2*ctrl_amp_env, initial_state, dt, frequencies, number_channels


def test_PulseForm_initialisation():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = PulseForm(original_system, pulse_function)
    assert pulse_form.evolver == original_system.evolver
    assert np.array_equal(pulse_form.H0, original_system.H0)
    assert np.array_equal(pulse_form.Hs, original_system.Hs)
    assert pulse_form.hilbert_space == original_system.hilbert_space
    assert pulse_form.pulse_function == pulse_function
    assert pulse_form.appended == False

    pulse_form2 = PulseForm(original_system, pulse_function_append, append=True)
    assert pulse_form2.evolver == original_system.evolver
    assert np.array_equal(pulse_form2.H0, original_system.H0)
    assert np.array_equal(pulse_form2.Hs, original_system.Hs)
    assert pulse_form2.hilbert_space == original_system.hilbert_space
    assert pulse_form2.pulse_function == pulse_function_append
    assert pulse_form2.appended == True

def test_PulseForm_initialisation_from_QuantumSystem():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = original_system.pulse_form(pulse_function)
    assert pulse_form.evolver == original_system.evolver
    assert np.array_equal(pulse_form.H0, original_system.H0)
    assert np.array_equal(pulse_form.Hs, original_system.Hs)
    assert pulse_form.hilbert_space == original_system.hilbert_space
    assert pulse_form.pulse_function == pulse_function
    assert pulse_form.appended == False

    pulse_form2 = original_system.pulse_form(pulse_function_append, append=True)
    assert pulse_form2.evolver == original_system.evolver
    assert np.array_equal(pulse_form2.H0, original_system.H0)
    assert np.array_equal(pulse_form2.Hs, original_system.Hs)
    assert pulse_form2.hilbert_space == original_system.hilbert_space
    assert pulse_form2.pulse_function == pulse_function_append
    assert pulse_form2.appended == True

def test_pulse_form():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = PulseForm(original_system, pulse_function)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    output1 = pulse_form._pre_processing()
    output2 = original_system._pre_processing(*args)
    assert len(output1) == len(output2)
    for out1, out2 in zip(output1, output2):
        assert np.array_equal(out1, out2)
    
def test_pulse_form_append():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = PulseForm(original_system, pulse_function_append, append=True)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args1 = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    args2 = (2*ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    output1 = pulse_form._pre_processing(*args1)
    output2 = original_system._pre_processing(*args2)
    assert len(output1) == len(output2)
    for out1, out2 in zip(output1, output2):
        assert np.array_equal(out1, out2)

def test_pulse_function_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = PulseForm(original_system, pulse_function)
    try:
        pulse_form.pulse_function = lambda *x: x
    except AttributeError:
        pass
    else:
        raise AssertionError("PulseFrom.pulse_function should be read-only")

def test_appended_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    original_system = QuantumSystem(H0, Hs, hilbert_space)
    original_system.initialise_evolver()
    pulse_form = PulseForm(original_system, pulse_function)
    try:
        pulse_form.appended = True
    except AttributeError:
        pass
    else:
        raise AssertionError("PulseFrom.appended should be read-only")


"""
Defines classes for quantum systems
"""

from typing import Callable, Optional, Union

import numpy as np
import tensorflow as tf

from py_ste import get_unitary_evolver
from py_ste.evolvers import UnitaryEvolver

from .._hilbert_space import HilbertSpace
from ..pulses import compose_unpack

def generate_channel_couplings(number_channels: list[int]) -> np.ndarray[bool]:
    """Generates a boolean array indicating which channels couple to which
    drives.

    Parameters
    ----------
    number_chanels : list[int]
        A list with the nth entry indicating the number of channels that
        correspond the nth drive.

    Returns
    -------
    NDArray[Shape[``len(number_channels)``, total_number_channels], bool]
        An entry is `True` if the channel corresponds to the drive.
    """
    number_channels: np.ndarray[int] = np.array(number_channels)
    number_channels = number_channels.cumsum()
    channel_couplings: np.ndarray[bool] = np.zeros((len(number_channels),
                                                    number_channels[-1]),
                                                   dtype=bool)
    c_old = 0
    for i, c in enumerate(number_channels):
        channel_couplings[i, c_old: (c_old := c)] = True
    return channel_couplings

# Defining classes
class ExpValCustom:
    """
    A class implimenting :meth:`QuantumSystem.evolved_expectation_value()` with
    a `TensorFlow <https://www.tensorflow.org>`__ gradient.
    """

    system: "QuantumSystem"
    "The system in which to take the expectation value"
    
    initial_state: np.ndarray[complex]
    "The initial state for the integrator"
    
    dt: float
    "The integrator time step"
    
    observable: np.ndarray[complex]
    "The observable to take the expectation value of"
    
    def __init__(self,
                 system: "QuantumSystem",
                 initial_state: np.ndarray[complex],
                 dt: float,
                 observable: np.ndarray[complex]):
        """
        Initialises the class with the `system` in which to take the expectation
        value and the `observable` in the `system` to take the expectation value
        of. Additionally, the initial state before evolution and
        the integrator time step are specified.

        Parameters
        ----------
        system : QuantumSystem
            The system in which to take the expectation value
        initial_state : NDArray[Shape[``system.state_shape``], complex]
            The initial state for the integrator
        dt : float
            The integrator time step
        observable : NDArray[Shape[``system.dim``, ``system.dim``], complex])
            The observable to take the expectation value of
        """
        self.system = system
        self.initial_state = initial_state
        self.dt = dt
        self.observable = observable
    @tf.custom_gradient
    def run(self, ctrl_amp):
        """
        Computes the expectation value of the :attr:`observable` in the
        :attr:`system` with respect to the :attr:`initial_state` evolved under
        the Hamiltonian generated by the specified control amplitudes.

        Parameters
        ----------
        ctrl_amp : tf.Tensor[Shape[n_time_steps, ``system.n_ctrl``], tf.complex128]
            The control amplitudes

        Returns
        -------
        tf.Tensor[Shape[], tf.complex128]
            The expectation value

        Note
        ----
        This function is differentiable using TensorFlow's ``tf.GradientTape``.
        """
        E, self.switching_function = self.system.evolver.switching_function(ctrl_amp.numpy().copy(), self.initial_state, self.dt, self.observable)
        self.switching_function = tf.constant(self.switching_function*self.dt, dtype=tf.complex128)
        def grad(upstream) -> tf.Tensor: return self.switching_function
        return tf.math.real(tf.constant(E, dtype=tf.complex128)), grad # @tf.custom_gradient removes the second returned variable

class QuantumSystem:
    """
    A class storing the properties of a quantum system.
    """
    _evolver: Optional[UnitaryEvolver] = None
    "The integrator used for time evolutions of the system."
    
    _hilbert_space: HilbertSpace
    "The Hilbert space of the system"
    
    _H0: np.ndarray
    "The systems drift Hamiltonian as a :attr:`dim` x :attr:`dim` matrix."
    
    _Hs: np.ndarray
    """
    An array of the system's control Hamiltonians with shape
    (:attr:`n_ctrl`, :attr:`dim`, :attr:`dim`).
    """
    
    _graph_processing: Callable[..., tuple]
    "A Tensorflow graph of :attr:`_processing()`"
    
    _processing: Callable[..., tuple]
    """
    Executes :meth:`_pre_processing()` followed by
    :meth:`_envolope_processing()` eagerly (i.e. without using a TensorFlow
    graph). Nonetheless, :meth:`_eager_processing()` is still auto
    differentiable.

    Parameters
    ----------
    *args
        The placeholder parameters. See ``_systems.pyi`` for actual parameters.
        Each child class that implements a new :meth:`_pre_processing()` should
        implement a ``.pyi`` file to document the parameters for this function:
        the same parameters as passed to :meth:`_pre_processing()`.

    Returns
    -------
    tuple[tf.Tensor[Shape[n_time_steps, :attr:`n_ctrl`], complex], tf.Tensor[Shape[:attr:`state_shape`], complex], tf.Tensor[Shape[], float]]
        A tuple of:
        1. Control amplitudes
        2. Initial state
        3. Integrator time step
    """
    
    _using_graph: bool
    """
    Whether to use TensorFlow graphs during computation. Using a TensorFlow
    graph will increase the speed of computation. However, you have to be
    careful that function parameters have not been baked into the graph leading
    to unexpected behaviour.
    """
    
    def __init__(self,
                 H0: np.ndarray[complex],
                 Hs: np.ndarray[complex],
                 hilbert_space: HilbertSpace,
                 use_graph: bool = True):
        """
        Initialises a new :class:`QuantumSystem`.

        Parameters
        ----------
        H0 : NDArray[Shape[:attr:`dim`, :attr:`dim`], complex]
            The systems drift Hamiltonian
        Hs : NDArray[Shape[:attr:`n_ctrl`, :attr:`dim`, :attr:`dim`], complex] | NDArray[Shape[:attr:`n_ctrl` * :attr:`dim`, :attr:`dim`], complex]
            The systems control Hamiltonians either as an array of control
            Hamiltonians or the control Hamiltonians stacked along the first
            axis.
        hilbert_space : HilbertSpace
            The Hilbert space of the system
        use_graph : bool
            Whether to use `TensorFlow <https://www.tensorflow.org>`__ graphs
            during computation, by default ``True``
        """
        self._hilbert_space: HilbertSpace = hilbert_space
        self._H0 = np.array(H0)
        self._H0.flags.writeable = False
        Hs = np.array(Hs)
        self._Hs = Hs if Hs.ndim == 2 else Hs.reshape(-1, self.dim)
        self._Hs.flags.writeable = False
        self._graph_processing = tf.function(self._traceable_eager_processing, autograph=False)
        self.using_graph = use_graph
            
    def __del__(self):
        # to force clear up of tracing
        del self._graph_processing
        del self._processing
    @property
    def using_graph(self) -> bool:
        """
        Whether to use `TensorFlow <https://www.tensorflow.org>`__ graphs during
        computation. Using a `TensorFlow <https://www.tensorflow.org>`__ graph
        will increase the speed of computation. However, you have to be careful
        that function parameters have not been baked into the graph leading to
        unexpected behaviour.
        """
        return self._using_graph
    @using_graph.setter
    def using_graph(self, value: bool):
        self._using_graph = value
        self._processing = self._graph_processing if value else self._eager_processing
    @property
    def hilbert_space(self) -> HilbertSpace:
        "The Hilbert space of the system"
        return self._hilbert_space
    @property
    def H0(self) -> np.ndarray[complex]:
        """
        The systems drift Hamiltonian as a :attr:`dim` x :attr:`dim` matrix.

        See Also
        --------
        :attr:`Hs`
        """
        return self._H0
    @property
    def Hs(self) -> np.ndarray[complex]:
        """
        An array of the system's control Hamiltonians with shape
        (:attr:`n_ctrl`, :attr:`dim`, :attr:`dim`).

        See Also
        --------
        :attr:`H0`
        """
        return self._Hs.reshape((-1, self.dim, self.dim))
    @property
    def dim(self) -> int:
        """
        The dimension of states in the quantum system.

        See Also
        --------
        :attr:`state_shape`
        """
        return self._hilbert_space.dim
    @property
    def state_shape(self) -> tuple[int]:
        """
        The shape of the states in the system.

        See Also
        --------
        :attr:`dim`
        """
        return (self.dim,)
    @property
    def n_ctrl(self) -> int:
        """
        The number of control Hamiltonians.
        """
        return len(self.Hs)
    @property
    def evolver(self) -> UnitaryEvolver:
        """
        The integrator used for time evolutions of the system.

        Note
        ----
        The `evolver` can take a while to initialise and so is not initialised
        until `evolver` is is first used or when :meth:`initialise_evolver()` is
        called. Using `evolver` before calling :meth:`initialise_evolver()`
        initialises the `evolver` with the default parameters of
        :meth:`initialise_evolver()`.
        """
        if self._evolver is None:
            self.initialise_evolver()
        return self._evolver
    def initialise_evolver(self,
                           sparse: bool = False,
                           force_dynamic: bool = False):
        """
        Initialises :attr:`evolver` with an evolver from
        `PySTE <https://PySTE.readthedocs.io>`__.
        `PySTE <https://PySTE.readthedocs.io>`__ is Python
        wrapper around the C++ header-only library
        `Suzuki-Trotter-Evolver <https://Suzuki-Trotter-Evolver.readthedocs.io>`__:
        a fast Schrödinger solver utilising the first-order Suzuki-Trotter
        expansion.

        Warning
        -------
        This can take a very long time to execute, especially for large Hilbert
        space dimensions. If you plan to evolve the same quantum system many
        times we recommended pickling the :attr:`evolver`.

        Parameters
        ----------
        sparse : bool
            Whether to use sparse or dense matrices during integration.
            To make a decision on whether sparse or dense matrices are likely to
            lead to faster integration you can consult the benchmarks at
            https://PySTE.readthedocs.io/en/latest/benchmarks.
        force_dynamic : bool
            Whether to force `PySTE <https://PySTE.readthedocs.io>`__ to use a
            dynamic evolver.
            
            Note
            ----
            `PySTE <https://PySTE.readthedocs.io>`__ has precompiled evolvers
            for specific Hilbert space dimensions and numbers of control
            Hamiltonians. When these cannot be found
            `PySTE <https://PySTE.readthedocs.io>`__ uses less efficient
            evolvers with the Hilbert space dimension and the number of controls
            determined dynamically at runtime.
        """
        self._evolver = get_unitary_evolver(self.H0, self._Hs, sparse, force_dynamic)
    def _H(self, ctrl_amp: np.ndarray[float]) -> np.ndarray[complex]:
        """
        Computes the system Hamiltonian for the specified control amplitudes.

        Parameters
        ----------
        ctrl_amp : NDArray[Shape[s := Any_Shape, :attr:`n_ctrl`], float]
            The control amplitudes (stored in the last axis). The prior axes
            allow for multiple sets of control amplitudes to be passed and the
            Hamiltonian for each computed.

        Returns
        -------
        NDArray[Shape[s, :attr:`dim`, :attr:`dim`], complex]
            The system's Hamiltonian (stored in the last two axes).
        """
        return self._H0 + np.einsum("...i,ijk->...jk", ctrl_amp, self.Hs)
    def H(self,
          ctrl_amp: Union[np.ndarray[float], np.ndarray[Callable[[float], np.ndarray[float]]], Callable[[float], np.ndarray[float]]]
         ) -> Union[np.ndarray[complex], Callable[[float], np.ndarray[complex]]]:
        """
        Computes the system Hamiltonian for the specified control amplitudes.

        Parameters
        ----------
        ctrl_amp : NDArray[Shape[s := Any_Shape, :attr:`n_ctrl`], float | Callable[[float], np.ndarray[float]]] | Callable[[float], NDArray[Shape[:attr:`n_ctrl`], float]
            The control amplitudes (stored in the last axis). The prior axes
            allow for multiple sets of control amplitudes to be passed and the
            Hamiltonian for each computed. The control amplitudes can be passed
            as ``np.ndarray[float]`` to compute the system Hamiltonian for a
            specific value of the control ampltiudes. Alternatively,
            time-dependent control amplitudes can be passed.
            ``np.ndarray[Callable[[float], np.ndarray[float]]]`` can be passed
            where each element is a function of time. Alternatively, a function
            of time  that returns the control amplitudes can be passed as
            ``Callable[[float], NDArray[Shape[:attr:`n_ctrl`], float]``. These
            will generate a time-dependent Hamiltonian: a function that takes a
            single parameter (time) and returns the Hamiltonian at this time.

        Returns
        -------
        NDArray[Shape[s, :attr:`dim`, :attr:`dim`], complex] | NDArray[Shape[s], Callable[[float], np.ndarray[complex]]]]
            Either the systems Hamiltonian stored in the last two axes (if
            specific control amplitudes were passed) or a collection of
            time-dependent Hamiltonians (if time-dependent controls were
            passed).
        """
        if callable(ctrl_amp):
            return lambda t: self._H(ctrl_amp(t))
        ctrl_amp = np.array(ctrl_amp)
        if ctrl_amp.dtype == object:
            return lambda t: self._H([a(t) for a in ctrl_amp])
        return self._H(ctrl_amp)
    def _pre_processing(self, *args):
        """
        When calling any evolution method (listed in the
        :ref:`See also section <pre_processing_see_also>`)
        :meth:`_pre_processing()` is executed on the arguments before the
        control amplitudes are modulated by the frequencies (during
        :meth:`_envolope_processing()`) and then finally the modulated control
        amplitudes are used by the evolution method.

        :meth:`_pre_processing()` should be overridden to produce desired pulse
        shapes. You can either override :meth:`_pre_processing()` directly by
        creating a child class, or you can use :meth:`pulse_form()`.

        For :meth:`gradient()` to function correctly :meth:`_pre_processing()`
        should be written in `TensorFlow <https://www.tensorflow.org>`__.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], tf.Tensor[Shape[:attr:`state_shape`], tf.complex128], float, tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], list[int]]
            A tuple of
            1. The control amplitude envolopes
            2. The initial state
            3. The integrator time step
            4. The frequencies to modulate the control amplitude envolopes with
            5. A list of the number of channels for each control Hamiltonian

            Warning
            -------
            The number of channels for each control Hamiltonian must be stored
            as a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.


        .. _pre_processing_see_also:
        
        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        * :meth:`get_driving_pulses()`
        * :meth:`gradient()`
        """
        return args
    def _envolope_processing(self,
                             ctrl_amp,
                             dt: float,
                             frequencies,
                             number_channels: list[int]
                            ) -> tuple:
        """
        When calling any evolution method (listed in the
        :ref:`See also section <envolope_processing_see_also>` section)
        :meth:`_pre_processing()` is executed on the arguements before the
        control amplitudes are modulated by the frequencies during
        :meth:`_envolope_processing()` and then finally the modulated control
        amplitudes are used by the evolution method.

        Parameters
        ----------
        ctrl_amp : tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128]
            The envolope control amplitudes
        dt : float
            The itegration time step
        frequencies : tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128]
            The frequencies to modulate the control amplitudes with
        number_channels : list[int]
            The number of channels associated with each control Hamiltonian

            Warning
            -------
            This must be a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.

        Returns
        -------
        tf.Tensor[Shape[n_time_steps, :attr:`n_ctrl`], tf.complex128]
            The modulated control amplitudes


        .. _envolope_processing_see_also:
        
        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        * :meth:`get_driving_pulses()`
        * :meth:`gradient()`
        """
        dt = tf.cast(dt, dtype=tf.complex128)
        frequencies = tf.cast(frequencies, dtype=tf.complex128)
        ctrl_amp = tf.cast(ctrl_amp, dtype=tf.complex128)
        channel_couplings = tf.constant(generate_channel_couplings(number_channels), dtype=tf.complex128)
        x = tf.exp(tf.einsum("i,j->ij", dt*tf.cast(tf.range(tf.shape(ctrl_amp)[0]), dtype=tf.complex128), -1j*frequencies))
        ctrl_amp = tf.cast(tf.math.real(tf.einsum("tc,tc,dc->td", x, ctrl_amp, channel_couplings)), dtype=tf.complex128)
        return ctrl_amp
    def propagate(self, *args) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.propagate()`
        from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        NDArray[Shape[:attr:`state_shape`], complex]
            The final state

        See Also
        --------
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        """
        ctrl_amp, initial_state, dt = self.get_driving_pulses(*args)
        return self.evolver.propagate(ctrl_amp, initial_state, dt)
    def propagate_collection(self, *args) -> np.ndarray[complex]:
        """
        Evolves a collection of state vectors under the time-dependent
        Hamiltonian defined by the control amplitudes using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.propagate_collection()`
        from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        NDArray[Shape[n_states, :attr:`state_shape`], complex]
            The final state

        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_all()`
        """
        ctrl_amp, initial_state, dt = self.get_driving_pulses(*args)
        return self.evolver.propagate_collection(ctrl_amp, initial_state, dt)
    def propagate_all(self, *args) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.propagate_all()`
        from `PySTE <https://PySTE.readthedocs.io>`__ and returns the state at
        each time-step.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        NDArray[Shape[n_time_steps+1, :attr:`state_shape`], complex]
            The state at each integrator time step (including the initial
            state).

        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        """
        ctrl_amp, initial_state, dt = self.get_driving_pulses(*args)
        return self.evolver.propagate_all(ctrl_amp, initial_state, dt)
    def evolved_expectation_value(self, *args) -> complex:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the final state using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.evolved_expectation_value()`
        from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        ``*args[:-1]``
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.
        ``args[-1]`` : NDArray[Shape[:attr:`dim`, :attr:`dim`], complex]
            The observable to take the expectation value of.


        Returns
        -------
        complex
            The expectation value.

        See Also
        --------
        * :meth:`evolved_expectation_value_all()`
        * :meth:`gradient()`
        """
        observable: np.ndarray[complex] = args[-1]
        ctrl_amp, initial_state, dt = self.get_driving_pulses(*args[:-1])
        return self.evolver.evolved_expectation_value(ctrl_amp,
                                                      initial_state,
                                                      dt,
                                                      observable)
    def evolved_expectation_value_all(self, *args) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the state at each time-step using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.evolved_expectation_value_all()`
        from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        ``*args[:-1]``
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.
        ``args[-1]`` : NDArray[Shape[:attr:`dim`, :attr:`dim`], complex]
            The observable to take the expectation value of.

        Returns
        -------
        NDArray[Shape[n_time_steps+1], complex]
            The state at each integrator time step (including the initial
            state).

        See Also
        --------
        * :meth:`evolved_expectation_value()`
        * :meth:`gradient()`
        """
        observable: np.ndarray[complex] = args[-1]
        ctrl_amp, initial_state, dt = self.get_driving_pulses(*args[:-1])
        return self.evolver.evolved_expectation_value_all(ctrl_amp,
                                                          initial_state,
                                                          dt,
                                                          observable)
    def get_driving_pulses(self, *args) -> tuple[np.ndarray[complex], np.ndarray[complex], float]:
        """
        When calling any evolution method (listed in the
        :ref:`See also section <get_driving_pulses_see_also>`) :meth:`get_driving_pulses()`
        is executed on the arguements before the evolution method.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        tuple[NDArray[Shape[n_time_steps, :attr:`n_ctrl`], complex], NDArray[Shape[:attr:`state_shape`], complex], float]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step


        .. _get_driving_pulses_see_also:
        
        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        * :meth:`gradient()`
        """
        ctrl_amp, initial_state, dt = self._processing(*args)
        try: ctrl_amp: np.ndarray[complex] = ctrl_amp.numpy()
        except: pass
        try:  initial_state: np.ndarray[complex] = initial_state.numpy().flatten()
        except: pass
        try: dt = dt.numpy()
        except: pass
        dt = float(dt.real)
        return ctrl_amp, initial_state, dt
    def _eager_processing(self, *args) -> tuple:
        """
        Executes :meth:`_pre_processing()` followed by
        :meth:`_envolope_processing()` eagerly (i.e. without using a
        `TensorFlow <https://www.tensorflow.org>`__ graph). Nonetheless,
        :meth:`_eager_processing()` is still auto differentiable.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, :attr:`n_ctrl`], complex], tf.Tensor[Shape[:attr:`state_shape`], complex], tf.Tensor[Shape[], float]]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step
        """
        ctrl_amp, initial_state, dt, frequencies, number_channels = self._pre_processing(*args)
        return self._envolope_processing(ctrl_amp, dt, frequencies, list(number_channels)), initial_state, dt
    def _traceable_eager_processing(self, *args) -> tuple:
        """
        A function that will be traced by
        `TensorFlow <https://www.tensorflow.org>`__ to produce a graph of
        :meth:`_pre_processing()` followed by :meth:`_envolope_processing()`.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, :attr:`n_ctrl`], complex], tf.Tensor[Shape[:attr:`state_shape`], complex], tf.Tensor[Shape[], float]]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step
        """
        print("Tracing control amplitude graph.")
        return self._eager_processing(*args)
    def gradient(self, *args) -> tuple[float, np.ndarray[float]]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the final state and then computes the
        gradient of the final state with respect to the first argument
        (``args[0]``) using
        :meth:`~py_ste.evolvers.DenseUnitaryEvolver.switching_function()`
        from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        ``*args[:-1]``
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.
        ``args[-1]`` : NDArray[Shape[:attr:`dim`, :attr:`dim`], complex]
            The observable to take the expectation value of.

        Returns
        -------
        tuple[complex, NDArray[Shape[n_parameters], float]]
            A tuple of the expectation value and the gradient.

        See Also
        --------
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        """
        cost: np.ndarray[complex] = args[-1]
        args = list(args[:-1])
        args[0] = tf.constant(args[0], dtype=tf.float64)             
        with tf.GradientTape(persistent=False,
                             watch_accessed_variables=False
                            ) as tape:
            tape.watch(args[0])
            ctrl_amp, initial_state, dt = self._processing(*args)
            try: initial_state: np.ndarray[complex] = initial_state.numpy().flatten()
            except: pass
            try: dt = dt.numpy()
            except: pass
            dt = float(dt.real)
            E = ExpValCustom(self, initial_state, dt, cost).run(ctrl_amp)
        grad = tape.gradient(E, args[0])
        del tape
        
        grad = tf.convert_to_tensor(grad)
        grad: np.ndarray[float] = grad.numpy()
        grad = grad.real
        return E.numpy(), grad
    def pulse_form(self,
                   pulse_function: Callable,
                   append: bool = False,                  
                  ) -> "PulseForm":
        """
        Initialises a new :class:`QuantumSystem` in which
        :meth:`_pre_processing()` corresponds to executing ``pulse_function()``
        and piping the output into the previous definition of
        :meth:`_pre_processing()`.

        Parameters
        ----------
        pulse_function : Callable
            The function to compose with :meth:`_pre_processing()`.

        Returns
        -------
        PulseForm
            The new :class:`QuantumSystem`
        """
        return PulseForm(self, pulse_function, append)

class TransformedSystem(QuantumSystem):
    """
    A base class for representing a transformation on a :class:`qugrad.QuantumSystem`.
    """

    _original_system: QuantumSystem
    "The system that was transformed into this system"
    
    _base_system: QuantumSystem
    """
    The system before any transformations were applied. That is `_base_system`
    is the recursive :attr:`original_system`
    (``original_system.original_system.original_system....``) until
    :attr:`original_system` is no longer a :class:`TransformedSystem`.
    """
    
    def __init__(self,
                 original_system: QuantumSystem,
                 H0: np.ndarray[complex],
                 Hs: Union[np.ndarray[complex], np.ndarray[complex]],
                 hilbert_space: HilbertSpace):
        """
        Performs a transformation on a :class:`qugrad.QuantumSystem`.

        Parameters
        ----------
        original_system: QuantumSystem
            The system to be transformed into this system
        H0: NDArray[Shape[:attr:`dim`, :attr:`dim`], complex]
            The new drift Hamiltonian
        Hs: NDArray[Shape[":attr:`n_ctrl`, :attr:`dim`, :attr:`dim`"], complex] | NDArray[Shape[:attr:`n_ctrl` * :attr:`dim`, :attr:`dim`], complex]
            The new control Hamiltonians either as an array of control
            Hamiltonians or the control Hamiltonians stacked along the first
            axis.
        hilbert_space: HilbertSpace
            The new Hilbert space of the system
        """
        self._original_system = original_system
        if isinstance(original_system, TransformedSystem):
            self._base_system = original_system._base_system
        else:
            self._base_system = original_system
        super().__init__(H0, Hs, hilbert_space, self._base_system.using_graph)
    @property
    def original_system(self) -> QuantumSystem:
        "The system that was transformed into this system"
        return self._original_system
    @property
    def base_system(self) -> QuantumSystem:
        """
        The system before any transformations were applied. That is
        :attr:`base_system` is the recursive :attr:`original_system`
        (``original_system.original_system.original_system....``) until
        :attr:`original_system` is no longer a :class:`TransformedSystem`.
        """
        return self._base_system
    def _pre_processing(self, *args) -> tuple:
        """
        When calling any evolution method (listed in the
        :ref:`See also section <TransformedSystem_pre_processing_see_also>` section)
        :meth:`_pre_processing()` is executed on the arguements before the
        control amplitudes are modulated by the frequencies (during
        :meth:`_envolope_processing()`) and then finally the modulated control
        amplitudes are used by the evolution method.

        This is a placeholder for ``original_system._pre_processing()``.

        Parameters
        ----------
        *args
            The placeholder parameters. See ``_systems.pyi`` for actual
            parameters. Each child class that implements a new
            :meth:`_pre_processing()` should implement a ``.pyi`` file to
            document the parameters for this function: the same parameters as
            passed to :meth:`_pre_processing()`.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], tf.Tensor[Shape[:attr:`state_shape`], tf.complex128], float, tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], list[int]]
            A tuple of
            1. The control amplitude envolopes
            2. The initial state
            3. The integrator time step
            4. The frequencies to modulate the control amplitude envolopes with
            5. A list of the number of channels for each control Hamiltonian

            Warning
            -------
            The number of channels for each control Hamiltonian must be stored
            as a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.


        .. _TransformedSystem_pre_processing_see_also:
        
        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        * :meth:`get_driving_pulses()`
        * :meth:`gradient()`
        """
        return self._original_system._pre_processing(*args)
    def _envolope_processing(self,
                             ctrl_amp,
                             dt: float,
                             frequencies,
                             number_channels: list[int]):
        """
        When calling any evolution method (listed in the
        :ref:`See also section <envolope_processing_see_also>` section) :meth:`_pre_processing()`
        is executed on the arguements before the control amplitudes are
        modulated by the frequencies during :meth:`_envolope_processing()` and
        then finally the modulated control amplitudes are used by the evolution
        method.

        Parameters
        ----------
        ctrl_amp : tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128]
            The envolope control amplitudes
        dt : float
            The itegration time step
        frequencies : tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128]
            The frequencies to modulate the control amplitudes with
        number_channels : list[int]
            The number of channels associated with each control Hamiltonian

            Warning
            -------
            This must be a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.

        Returns
        -------
        tf.Tensor[Shape[n_time_steps,:attr:`n_ctrl`], tf.complex128]
            The modulated control amplitudes


        .. _envolope_processing_see_also:
        
        See Also
        --------
        * :meth:`propagate()`
        * :meth:`propagate_collection()`
        * :meth:`propagate_all()`
        * :meth:`evolved_expectation_value()`
        * :meth:`evolved_expectation_value_all()`
        * :meth:`get_driving_pulses()`
        * :meth:`gradient()`
        """
        return self._original_system._envolope_processing(ctrl_amp,
                                                          dt,
                                                          frequencies,
                                                          number_channels)

class PulseForm(TransformedSystem):
    """
    A transformed :class:`qugrad.QuantumSystem` in which :meth:`_pre_processing()`
    has been composed with another pre processing function.
    """

    _pulse_function: Callable
    "The function composed with ``original_system._pre_processing()``"
    
    _appended: bool
    """
    Whether the :attr:`pulse_function` was prepended
    (``original_system._pre_processing(*pulse_function())``) or appended
    (``pulse_function(*original_system._pre_processing())``)
    """
        
    def __init__(self,
                 original_system: QuantumSystem,
                 pulse_function: Callable,
                 append: bool = False):
        """
        Initialises a new :class:`qugrad.QuantumSystem` in which :meth:`_pre_processing()`
        corresponds to running ``pulse_function()`` and piping the output into
        ``original_system._pre_processing()``.

        Parameters
        ----------
        original_system : QuantumSystem
            The system that was transformed into this system
        pulse_function : Callable
            The function to compose with :meth:`_pre_processing()`.
        append : bool
            Whether to prepend
            (``original_system._pre_processing(*pulse_function())``) or append
            (``pulse_function(*original_system._pre_processing())``)
            ``pulse_function``.
        """
        super().__init__(original_system,
                         original_system.H0,
                         original_system.Hs,
                         original_system.hilbert_space)
        self._evolver = original_system._evolver
        if append:
            self._pre_processing = compose_unpack(
                pulse_function,
                original_system._pre_processing
            )
        else:
            self._pre_processing = compose_unpack(
                original_system._pre_processing,
                pulse_function
            )
        self._pulse_function = pulse_function
        self._appended = append
    @property
    def pulse_function(self) -> Callable:
        "The function composed with ``original_system._pre_processing()``"
        return self._pulse_function
    @property
    def appended(self) -> bool:
        """
        Whether the :attr:`pulse_function` was prepended
        (``original_system._pre_processing(*pulse_function())``) or appended
        (``pulse_function(*original_system._pre_processing())``)
        """
        return self._appended


from qugrad import HilbertSpace
import numpy as np

def test_initialisation():
    """
    Test the initialisation of the HilbertSpace class.
    """
    basis = [0, 1, 5, 3]
    hilbert_space = HilbertSpace(basis)
    
    # Check that the basis is correctly set
    assert np.array_equal(hilbert_space.basis, basis)
    
    # Check that the inverse is correctly set
    x = hilbert_space.basis[hilbert_space.inverse[hilbert_space.basis]]
    assert np.array_equal(x, basis)
    
    # Check that the dimension is correct
    assert hilbert_space.dim == len(basis)

def test_basis_read_only():
    hilbert_space = HilbertSpace([0])
    try:
        hilbert_space.basis = np.array([0])
    except AttributeError:
        pass
    else:
        raise AssertionError("HilbertSpace.basis should be read-only")
    
    try:
        hilbert_space.basis[0] = 0
    except ValueError:
        pass
    else:
        raise AssertionError("HilbertSpace.basis should be read-only")

def test_inverse_read_only():
    hilbert_space = HilbertSpace([0])

    try:
        hilbert_space.inverse = np.array([0])
    except AttributeError:
        pass
    else:
        raise AssertionError("HilbertSpace.inverse should be read-only")
    
    try:
        hilbert_space.inverse[0] = 0
    except ValueError:
        pass
    else:
        raise AssertionError("HilbertSpace.inverse should be read-only")

def test_dim_read_only():
    hilbert_space = HilbertSpace([0])
    try:
        hilbert_space.dim = 1
    except AttributeError:
        pass
    else:
        raise AssertionError("HilbertSpace.dim should be read-only")

def test_iter():
    hilbert_space = HilbertSpace([0, 1, 5, 3])
    assert len(hilbert_space) == len(hilbert_space.basis)
    for n, label in enumerate(hilbert_space):
        assert label == hilbert_space.basis[n]
        assert label == hilbert_space[n]

def test_basis_vector(): 
    hilbert_space = HilbertSpace([0, 1, 5, 3])

    for n, label in enumerate(hilbert_space.basis):
        assert np.array_equal(hilbert_space.basis_vector(label),
                              np.array([0]*n+[1]+[0]*(len(hilbert_space)-n-1)))

def test_get_subspace():
    hilbert_space = HilbertSpace([0, 1, 5, 3])
    filter = np.array([True, False, True, False])
    subspace = hilbert_space.get_subspace(filter)

    assert isinstance(subspace, HilbertSpace)
    
    assert np.array_equal(subspace.basis, hilbert_space.basis[filter])
    
    x = subspace.basis[subspace.inverse[subspace.basis]]
    assert np.array_equal(x, subspace.basis)
    
    assert subspace.dim == len(subspace.basis)

def test_labels():
    hilbert_space = HilbertSpace([0, 1, 5, 3])
    labels = hilbert_space.labels()
    for n, label in enumerate(labels):
        assert label == f"|{str(hilbert_space[n])}⟩"

    some_labels = hilbert_space.labels([1, 5])

    assert np.array_equal(some_labels, ["|1⟩", "|5⟩"])

    single_label = hilbert_space.labels(1)
    assert single_label == "|1⟩"

def test_empty_HilbertSpace():
    hilbert_space = HilbertSpace([])
    assert len(hilbert_space) == 0
    assert len(hilbert_space.basis) == 0
    assert len(hilbert_space.inverse) == 0
    try:
        hilbert_space.basis_vector(0)
    except IndexError:
        pass
    else:
        raise AssertionError("HilbertSpace.basis_vector should raise IndexError for empty HilbertSpace")
    subspace = hilbert_space.get_subspace([])
    assert isinstance(subspace, HilbertSpace)
    assert len(subspace) == 0
    assert len(subspace.basis) == 0
    assert len(subspace.inverse) == 0
    assert len(hilbert_space.labels()) == 0

def test_equality():
    hilbert_space1 = HilbertSpace([0, 1, 5, 3])
    hilbert_space2 = HilbertSpace([0, 1, 5, 3])
    hilbert_space3 = HilbertSpace([0, 1, 2])
    class SubclassHilbertSpace(HilbertSpace):
        pass
    hilbert_space4 = SubclassHilbertSpace([0, 1, 5, 3])

    assert hilbert_space1 == hilbert_space2
    assert not (hilbert_space1 == hilbert_space3)
    assert not (hilbert_space1 == "not a HilbertSpace")
    assert not (hilbert_space1 == hilbert_space4)

    assert not(hilbert_space1 != hilbert_space2)
    assert hilbert_space1 != hilbert_space3
    assert hilbert_space1 != "not a HilbertSpace"
    assert hilbert_space1 != hilbert_space4


# examples/subclass.py

import numpy as np
import tensorflow as tf

from functools import reduce
from qugrad import QuantumSystem, HilbertSpace

def kron(*args) -> np.ndarray:
    """
    Kronecker product of multiple matrices.

    Parameters
    ----------
    args: list[NDArray]
        The matrices to take the Kronecker product of

    Returns
    -------
    NDArray
        The Kronecker product of the matrices
    """
    return reduce(np.kron, args)

class ExampleSubclass(QuantumSystem):
    """
    An example subclass of QuantumSystem.
    """  
    def __init__(self,
                 qubits: int,
                 use_graph: bool = True):
        """
        Initialises the ExampleSubclass.

        Parameters
        ----------
        qubits: int
            The number of qubits in the system
        use_graph: bool
            Whether to use TensorFlow graphs during computation.
        """
        self.qubits = qubits
        X = np.array([[0,  1],  # Pauli-X
                      [1,  0]])
        Z = np.array([[1,  0],  # Pauli-Z
                      [0, -1]])
        H0 = kron(*[Z]*qubits)
        Xn = lambda n: kron(np.identity(2**(n-1)),
                            X,
                            np.identity(2**(qubits-n)))
        Hs = [Xn(n) for n in range(1, qubits+1)]
        super().__init__(H0,
                         Hs,
                         HilbertSpace(np.arange(2**qubits)),
                         use_graph)

    def _pre_processing(self,
                        frequencies: np.ndarray,
                        amplitudes: np.ndarray,
                        T: float
                       ) -> tuple:
        """
        When calling any evolution method (listed in the See also section)
        `_pre_processing()` is executed on the arguements before the control
        amplitudes are modulated by the frequencies (during
        `_envolope_processing()`) and then finally the modulated control
        amplitudes are used by the evolution method.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], tf.Tensor[Shape[:attr:`state_shape`], tf.complex128], float, tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], list[int]]
            A tuple of
            1. The control amplitude envolopes
            2. The initial state
            3. The integrator time step
            4. The frequencies to modulate the control amplitude envolopes with
            5. A list of the number of channels for each control Hamiltonian

            Warning
            -------
            The number of channels for each control Hamiltonian must be stored
            as a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.


        See Also
        --------
        * `propagate()`
        * `propagate_collection()`
        * `propagate_all()`
        * `evolved_expectation_value()`
        * `evolved_expectation_value_all()`
        * `get_driving_pulses()`
        * `gradient()`
        """
        amplitudes = tf.cast(amplitudes, tf.complex128)
        amplitudes = tf.reshape(amplitudes, (1, self.qubits))
        n_time_steps = tf.cast(tf.math.ceil(T)*100, tf.int32)
        ctrl_amp = tf.broadcast_to(amplitudes, tf.stack([n_time_steps, self.qubits], axis=0))
        initial_state = self.hilbert_space.basis_vector(0)
        dt = T/tf.cast(n_time_steps, tf.float64)
        frequencies = tf.reshape(tf.cast(frequencies, dtype=tf.complex128), (self.qubits,))
        number_channels = [1]*self.qubits
        return super()._pre_processing(ctrl_amp, initial_state, dt, frequencies, number_channels)


import os
import sys
FILE_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(os.path.dirname(FILE_DIR), "examples"))


def test_Rabi_oscillation_optimisation():
    import Rabi_oscillation_optimisation

def test_subclass():
    import subclass
    subclass.ExampleSubclass(2).propagate([1, 2], [3, 4], 5.)


# examples/Rabi_oscillation_optimisation.py

import numpy as np
import tensorflow as tf
from scipy.optimize import minimize

from qugrad import QuantumSystem, HilbertSpace

# Constants
X = np.array([[0,  1],  # Pauli-X
              [1,  0]],
             dtype=np.complex128)
Z = np.array([[1,  0],  # Pauli-Z
              [0, -1]],
             dtype=np.complex128)

INITIAL_STATE = np.array([1, 0], dtype=np.complex128)  # |0>
N_TIME_STEPS = 1000
T = 10
OBSERVABLE = Z

# Hamiltonian
H0 = Z
Hs = [X]

# Define quantum syste
hilbert_space = HilbertSpace([0, 1])
device = QuantumSystem(H0, Hs, hilbert_space)

# Define pulse form
def Rabi_drive(frequency, amplitude):
    amplitude = tf.cast(amplitude, tf.complex128)
    ctrl_amp = amplitude*tf.ones((N_TIME_STEPS, 1), dtype=tf.complex128)
    initial_state = INITIAL_STATE
    dt = T/N_TIME_STEPS
    frequencies = tf.reshape(tf.cast(frequency, dtype=tf.complex128), (1,))
    number_channels = [1]
    return ctrl_amp, initial_state, dt, frequencies, number_channels

# SciPy minimization expects an array of variables so we will wrap our pulse
#   form in this upack function
def unpack_variables(x):
    frequency = x[0]
    amplitude = x[1]
    return frequency, amplitude

# Adding the pulses to the device
driven_device = device.pulse_form(Rabi_drive).pulse_form(unpack_variables)

# Initial parameters
x0 = [1,   # frequency
      0.1] # amplitude

# Expectation value before optimisation
initial_expectation_value = driven_device.evolved_expectation_value(x0,
                                                                    OBSERVABLE)

# Optimisation
result = minimize(driven_device.gradient,
                  x0=x0,
                  args=(OBSERVABLE,),
                  jac=True)

# Print results
print("Initial expectation value: ", initial_expectation_value)
print("Optimisation result:")
print(result)


"Utility functions for pulse generation and manipulation."

import typing

from typing import Callable

g_out = typing.TypeVar("g_out")
g_args = typing.TypeVar("g_args")
g_kwargs = typing.TypeVar("g_kwargs")
f_out = typing.TypeVar("f_out")
f_args = typing.TypeVar("f_args")
f_kwargs = typing.TypeVar("f_kwargs")

def compose(f: Callable[[g_out, f_args, f_kwargs], f_out],
            g: Callable[[g_args, g_kwargs], g_out],
            *args: f_args,
            **kwargs: f_kwargs
           ) -> Callable[[g_args, g_kwargs], f_out]:
    """
    Creates a function ``h`` given by the composition of `f` and `g`::

        h(...)=f(g(...))
        
    That is the output of `g` is piped into the first argument of `f`.

    Additionally, additionaly arguments and keyword arguments can be passed to
    `f` using `args` and `kwargs`::

        h(...)=f(g(...), *args, **kwargs)
        
    Explicitly ``h`` is defined as::

        h = lambda *a, **kw: f(g(*a, **kw), *args, **kwargs)
        
    Parameters
    ----------
    f : Callable[[g_out, f_args, f_kwargs], f_out]
        The second function to be called in the composition
    g : Callable[[g_args, g_kwargs], g_out]
        The first function to be called in the composition
    *args : f_args
        Additional arguments to pass to ``f``
    **kwargs : f_kwargs
        Additional keyword arguments to pass to ``f``

    Returns
    -------
    Callable[[g_args, g_kwargs], f_out]
        The composite function ``h``

    See Also
    --------
    :func:`compose_unpack()`
    """
    return lambda *a, **kw: f(g(*a, **kw), *args, **kwargs)

def compose_unpack(f: Callable[[g_out, f_args, f_kwargs], f_out],
                   g: Callable[[g_args, g_kwargs], g_out],
                   *args: f_args,
                   **kwargs: f_kwargs
                  ) -> Callable[[g_args, g_kwargs], f_out]:
    """
    Creates a function ``h`` given by the composition of `f` and ``*g``::
        
        h(...) = f(*g(...))

    That is the output of `g` is unpacked and piped into the arguments of `f`.

    Additionally, additionaly arguments and keyword arguments can be passed to
    `f` using `args` and `kwargs`::

        h(...) = f(*g(...), *args, **kwargs)

    Explicitly ``h`` is defined as::

        h = lambda *a, **kw: f(*g(*a, **kw), *args, **kwargs)

    Parameters
    ----------
    f : Callable[[g_out, f_args, f_kwargs], f_out]
        The second function to be called in the composition
    g : Callable[[g_args, g_kwargs], g_out]
        The first function to be called in the composition
    *args : f_args
        Additional arguments to pass to ``f``
    **kwargs : f_kwargs
        Additional keyword arguments to pass to ``f``

    Returns
    -------
    Callable[[g_args, g_kwargs], f_out]
        The composite function ``h``

    See Also
    --------
    :func:`compose()`
    """
    return lambda *a, **kw: f(*g(*a, **kw), *args, **kwargs)


# examples/subclass.pyi

import numpy as np

from qugrad import QuantumSystem
from typing import Callable

def kron(*args) -> np.ndarray:
    """
    Kronecker product of multiple matrices.

    Parameters
    ----------
    args: list[NDArray]
        The matrices to take the Kronecker product of

    Returns
    -------
    NDArray
        The Kronecker product of the matrices
    """
    ...

class ExampleSubclass(QuantumSystem):
    """
    An example subclass of QuantumSystem.
    """
    
    _processing: Callable[..., tuple]
    """
    Executes `_pre_processing()` followed by
    `_envolope_processing()` eagerly (i.e. without using a TensorFlow
    graph). Nonetheless, `_eager_processing()` is still auto
    differentiable.

    Parameters
    ----------
    frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
    amplitudes: NDArray[Shape[`qubits`], complex]
        The amplitude of to drive X on each of the qubits
    T: float
        The time to evolve the system for

    Returns
    -------
    tuple[tf.Tensor[Shape[n_time_steps, `n_ctrl`], complex], tf.Tensor[Shape[`state_shape`], complex], tf.Tensor[Shape[], float]]
        A tuple of:
        1. Control amplitudes
        2. Initial state
        3. Integrator time step
    """
    def __init__(self,
                 qubits: int,
                 use_graph: bool = True):
        """
        Initialises the ExampleSubclass.

        Parameters
        ----------
        qubits: int
            The number of qubits in the system
        use_graph: bool
            Whether to use TensorFlow graphs during computation.
        """
        ...
    def _pre_processing(self,
                        frequencies: np.ndarray,
                        amplitudes: np.ndarray,
                        T: float
                       ) -> tuple:
        """
        When calling any evolution method (listed in the See also section)
        `_pre_processing()` is executed on the arguements before the control
        amplitudes are modulated by the frequencies (during
        `_envolope_processing()`) and then finally the modulated control
        amplitudes are used by the evolution method.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], tf.Tensor[Shape[`state_shape`], tf.complex128], float, tf.Tensor[Shape[n_time_steps, total_n_channels], tf.complex128], list[int]]
            A tuple of
            1. The control amplitude envolopes
            2. The initial state
            3. The integrator time step
            4. The frequencies to modulate the control amplitude envolopes with
            5. A list of the number of channels for each control Hamiltonian

            Warning
            -------
            The number of channels for each control Hamiltonian must be stored
            as a ``list`` and not an ``NDArray`` or a
            `TensorFlow <https://www.tensorflow.org>`__ tensor.


        See Also
        --------
        * `propagate()`
        * `propagate_collection()`
        * `propagate_all()`
        * `evolved_expectation_value()`
        * `evolved_expectation_value_all()`
        * `get_driving_pulses()`
        * `gradient()`
        """
        ...
    def propagate(self,
                  ctrl_amp : np.ndarray[complex],
                  initial_state : np.ndarray[complex],
                  dt: float,
                  frequencies: np.ndarray[complex],
                  number_channels: list[int]
                 ) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes using `propagate()` from
        `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        NDArray[Shape[`state_shape`], complex]
            The final state

        See Also
        --------
        * `propagate_collection()`
        * `propagate_all()`
        """
        ...
    def propagate_collection(self,
                             ctrl_amp : np.ndarray[complex],
                             initial_states : np.ndarray[complex],
                             dt: float,
                             frequencies: np.ndarray[complex],
                             number_channels: list[int]
                            ) -> np.ndarray[complex]:
        """
        Evolves a collection of state vectors under the time-dependent
        Hamiltonian defined by the control amplitudes using
        `propagate_collection()` from `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        NDArray[Shape[n_states, `state_shape`], complex]
            The final state

        See Also
        --------
        * `propagate()`
        * `propagate_all()`
        """
        ...
    def propagate_all(self,
                      ctrl_amp : np.ndarray[complex],
                      initial_states : np.ndarray[complex],
                      dt: float,
                      frequencies: np.ndarray[complex],
                      number_channels: list[int]
                     ) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes using `propagate_all()` from
        `PySTE <https://PySTE.readthedocs.io>`__ and returns the state at each
        time-step.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        NDArray[Shape[n_time_steps+1, `state_shape`], complex]
            The state at each integrator time step (including the initial
            state).

        See Also
        --------
        * `propagate()`
        * `propagate_collection()`
        """
        ...
    def evolved_expectation_value(self,
                                  ctrl_amp : np.ndarray[complex],
                                  initial_state : np.ndarray[complex],
                                  dt: float,
                                  frequencies: np.ndarray[complex],
                                  number_channels: list[int],
                                  observable : np.ndarray[complex]
                                 ) -> complex:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the final state using
        `evolved_expectation_value()` from
        `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for
        observable : NDArray[Shape[`dim`, `dim`], complex]
            The observable to take the expectation value of.

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        complex
            The expectation value.

        See Also
        --------
        * `evolved_expectation_value_all()`
        * `gradient()`
        """
        ...
    def evolved_expectation_value_all(self,
                                      ctrl_amp : np.ndarray[complex],
                                      initial_state : np.ndarray[complex],
                                      dt: float,
                                      frequencies: np.ndarray[complex],
                                      number_channels: list[int],
                                      observable : np.ndarray[complex]
                                     ) -> np.ndarray[complex]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the state at each time-step using
        `evolved_expectation_value_all()` from
        `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for
        observable : NDArray[Shape[`dim`, `dim`], complex]
            The observable to take the expectation value of.

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        NDArray[Shape[n_time_steps+1], complex]
            The state at each integrator time step (including the initial
            state).

         See Also
        --------
        * `evolved_expectation_value()`
        * `gradient()`
        """
        ...
    def get_driving_pulses(self,
                           ctrl_amp : np.ndarray[complex],
                           initial_states : np.ndarray[complex],
                           dt: float,
                           frequencies: np.ndarray[complex],
                           number_channels: list[int]
                          ) -> tuple[np.ndarray[complex], np.ndarray[complex], float]:
        """
        When calling any evolution method (listed in the See also section`)
        `get_driving_pulses()` is executed on the arguements before the
        evolution method.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        tuple[NDArray[Shape[n_time_steps, `n_ctrl`], complex], NDArray[Shape[`state_shape`], complex], float]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step


        .. _get_driving_pulses_see_also:
        
        See Also
        --------
        * `propagate()`
        * `propagate_collection()`
        * `propagate_all()`
        * `evolved_expectation_value()`
        * `evolved_expectation_value_all()`
        * `gradient()`
        """
        ...
    def _eager_processing(self,
                          ctrl_amp : np.ndarray[complex],
                          initial_states : np.ndarray[complex],
                          dt: float,
                          frequencies: np.ndarray[complex],
                          number_channels: list[int]
                         ) -> tuple:
        """
        Executes `_pre_processing()` followed by
        `_envolope_processing()` eagerly (i.e. without using a
        `TensorFlow <https://www.tensorflow.org>`__ graph). Nonetheless,
        `_eager_processing()` is still auto differentiable.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, `n_ctrl`], complex], tf.Tensor[Shape[`state_shape`], complex], tf.Tensor[Shape[], float]]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step
        """
        ...
    def _traceable_eager_processing(self,
                                   ctrl_amp : np.ndarray[complex],
                                   initial_states : np.ndarray[complex],
                                   dt: float,
                                   frequencies: np.ndarray[complex],
                                   number_channels: list[int]
                                  ) -> tuple:
        """
        A function that will be traced by
        `TensorFlow <https://www.tensorflow.org>`__ to produce a graph of
        `_pre_processing()` followed by `_envolope_processing()`.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        tuple[tf.Tensor[Shape[n_time_steps, `n_ctrl`], complex], tf.Tensor[Shape[`state_shape`], complex], tf.Tensor[Shape[], float]]
            A tuple of:
            1. Control amplitudes
            2. Initial state
            3. Integrator time step
        """
        ...
    def gradient(self,
                 ctrl_amp : np.ndarray[complex],
                 initial_state : np.ndarray[complex],
                 dt: float,
                 frequencies: np.ndarray[complex],
                 number_channels: list[int],
                 observable : np.ndarray[complex]
                ) -> tuple[float, np.ndarray[float]]:
        """
        Evolves a state vector under the time-dependent Hamiltonian defined by
        the control amplitudes and computes the expectation value of a specified
        observable with respect to the final state and then computes the
        gradient of the final state with respect to the first argument
        (`args[0]`) using `switching_function()` from
        `PySTE <https://PySTE.readthedocs.io>`__.

        Parameters
        ----------
        frequencies: NDArray[Shape[`qubits`], float]
            The frequencies of the to drive X on each of the qubits
        amplitudes: NDArray[Shape[`qubits`], complex]
            The amplitude of to drive X on each of the qubits
        T: float
            The time to evolve the system for
        observable : NDArray[Shape[`dim`, `dim`], complex]
            The observable to take the expectation value of.

        Warning
        -------
        Keyword arguments are not supported.

        Returns
        -------
        tuple[complex, NDArray[Shape[n_parameters], float]]
            A tuple of the expectation value and the gradient.

        See Also
        --------
        * `evolved_expectation_value()`
        * `evolved_expectation_value_all()`
        """
        ...


import numpy as np
import tensorflow as tf
import py_ste
from py_ste import evolvers

from qugrad import QuantumSystem, HilbertSpace
from qugrad.systems._systems import generate_channel_couplings, ExpValCustom
from .pauli_matrices import X, Y, Z

def test_generate_channel_couplings():
    number_channels = [1, 4, 2, 3]
    channel_couplings = generate_channel_couplings(number_channels)
    assert np.array_equal(channel_couplings.shape,
                          (len(number_channels), sum(number_channels)))
    assert channel_couplings.dtype == bool
    assert np.array_equal(channel_couplings,
        [[True,  False, False, False, False, False, False, False, False, False],
         [False, True,  True,  True,  True,  False, False, False, False, False],
         [False, False, False, False, False, True,  True,  False, False, False],
         [False, False, False, False, False, False, False, True,  True,  True]])
    

def test_QuantumSystem_initialisation():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    assert np.array_equal(quantum_system.H0, H0)
    assert np.array_equal(quantum_system.Hs, Hs)
    assert quantum_system.hilbert_space == hilbert_space
    assert quantum_system.using_graph == True
    assert quantum_system.dim == 2
    assert np.array_equal(quantum_system.state_shape, (2,))
    assert quantum_system.n_ctrl == 2
    assert quantum_system._evolver is None

    quantum_system = QuantumSystem(H0, Hs, hilbert_space, use_graph=False)
    assert np.array_equal(quantum_system.H0, H0)
    assert np.array_equal(quantum_system.Hs, Hs)
    assert quantum_system.hilbert_space == hilbert_space
    assert quantum_system.using_graph == False
    assert quantum_system.dim == 2
    assert np.array_equal(quantum_system.state_shape, (2,))
    assert quantum_system.n_ctrl == 2
    assert quantum_system._evolver is None

        

def test_hilbert_space_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.hilbert_space = HilbertSpace([0, 1, 2])
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.hilbert_space should be read-only")

def test_H0_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.H0 = X
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.H0 should be read-only")

    try:
        quantum_system.H0[0, 0] = 0
    except ValueError:
        pass
    else:
        raise AssertionError("QuantumSystem.H0 should be read-only")

def test_Hs_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.Hs = np.array([X, Z])
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.Hs should be read-only")

    try:
        quantum_system.Hs[0, 0, 0] = 0
    except ValueError:
        pass
    else:
        raise AssertionError("QuantumSystem.Hs should be read-only")

def test_dim_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.dim = 3
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.dim should be read-only")

def test_state_shape_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.state_shape = (3,)
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.state_shape should be read-only")
    try:
        quantum_system.state_shape[0] = 0
    except TypeError:
        pass
    else:
        raise AssertionError("QuantumSystem.state_shape should be read-only")

def test_n_ctrl_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.n_ctrl = 3
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.n_ctrl should be read-only")

def test_evolver_initialisation():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    quantum_system.evolver
    assert isinstance(quantum_system.evolver, evolvers.DenseUnitaryEvolver)

    quantum_system2 = QuantumSystem(H0, Hs, hilbert_space)
    quantum_system2.initialise_evolver()
    assert isinstance(quantum_system.evolver, evolvers.DenseUnitaryEvolver)

    quantum_system.initialise_evolver(sparse=True)
    assert isinstance(quantum_system.evolver, evolvers.SparseUnitaryEvolver)

    quantum_system.initialise_evolver(force_dynamic=True)
    assert quantum_system.evolver.__class__ ==  evolvers.DenseUnitaryEvolver

    quantum_system.initialise_evolver(sparse=True, force_dynamic=True)
    assert quantum_system.evolver.__class__ ==  evolvers.SparseUnitaryEvolver    

def test_evolver_read_only():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    try:
        quantum_system.evolver = py_ste.get_unitary_evolver(H0, Hs, sparse=True)
    except AttributeError:
        pass
    else:
        raise AssertionError("QuantumSystem.evolver should be read-only")

def test_H_static():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    np.array_equal(quantum_system.H([1, 2]), Z+X+2*Y)
    np.array_equal(quantum_system.H([[1, 2], [3, 4]]), [Z+X+2*Y, Z+3*X+4*Y])

def test_H_dynamic():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    pulse = lambda t: np.array([np.sin(t), np.cos(t)])
    H = quantum_system.H(pulse)
    assert np.array_equal(H(np.e), Z+pulse(np.e)[0]*X+pulse(np.e)[1]*Y)

    pulse = np.array([lambda t: np.sin(t), lambda t: np.cos(t)])
    H = quantum_system.H(pulse)
    assert np.array_equal(H(np.e), Z+pulse[0](np.e)*X+pulse[1](np.e)*Y)

def test_pre_processing():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    assert np.array_equal(quantum_system._pre_processing(12, "5"), (12, "5"))

def test_evnvolope_processing():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    ctrl_amp = quantum_system._envolope_processing(ctrl_amp_env,
                                                   dt,
                                                   frequencies,
                                                   number_channels)
    ctrl_amp2 = \
        np.array([[1,                   4+8],
                  [2*np.cos(0.5*dt),    7+9*np.cos(2.5*dt)+10*np.sin(2.5*dt)],
                  [3*np.sin(0.5*2*dt), 11*np.sin(2.5*2*dt)],
                  [np.cos(0.5*3*dt),    1+np.cos(2.5*3*dt)]])
    assert np.allclose(ctrl_amp, ctrl_amp2)

def test_processing():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    out = quantum_system._processing(ctrl_amp_env,
                                     initial_state,
                                     dt,
                                     frequencies,
                                     number_channels)
    assert np.array_equal(out[0],
                          quantum_system._envolope_processing(ctrl_amp_env,
                                                              dt,
                                                              frequencies,
                                                              number_channels))
    assert np.array_equal(out[1], initial_state)
    assert out[2] == dt

def test_using_graph():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    out1 = quantum_system._processing(*args)
    assert quantum_system._processing == quantum_system._graph_processing
    quantum_system.using_graph = False
    assert quantum_system._processing == quantum_system._eager_processing
    out2 = quantum_system._processing(*args)
    assert np.array_equal(out1[0], out2[0])
    assert np.array_equal(out1[1], out2[1])
    assert out1[2] == out2[2]
    quantum_system2 = QuantumSystem(H0, Hs, hilbert_space, use_graph=False)
    assert quantum_system2._processing == quantum_system2._eager_processing
    out3 = quantum_system2._processing(*args)
    assert np.array_equal(out1[0], out3[0])
    assert np.array_equal(out1[1], out3[1])
    assert out1[2] == out3[2]
    

def test_get_driving_pulses():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0])
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    out1 = quantum_system.get_driving_pulses(*args)
    out2 = quantum_system._processing(*args)
    assert np.array_equal(out1[0], out2[0])
    assert np.array_equal(out1[1], out2[1])
    assert out1[2] == out2[2]

def test_propagate():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    state = quantum_system.propagate(*args)
    py_ste_args = quantum_system.get_driving_pulses(*args)
    evolver = py_ste.get_unitary_evolver(H0, Hs)
    assert np.array_equal(state, evolver.propagate(*py_ste_args))

def test_propagate_collection():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([[1, 0],
                              [0, 1]],
                             dtype=complex)
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    state = quantum_system.propagate_collection(*args)
    py_ste_args = quantum_system.get_driving_pulses(*args)
    evolver = py_ste.get_unitary_evolver(H0, Hs)
    assert np.array_equal(state, evolver.propagate_collection(*py_ste_args))

def test_propagate_all():
    H0 = Z
    Hs = [X, Y]
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    state = quantum_system.propagate_all(*args)
    py_ste_args = quantum_system.get_driving_pulses(*args)
    evolver = py_ste.get_unitary_evolver(H0, Hs)
    assert np.array_equal(state, evolver.propagate_all(*py_ste_args))

def test_evolved_expectation_value():
    H0 = Z
    Hs = [X, Y]
    O = X+Y+Z
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    value = quantum_system.evolved_expectation_value(*args, O)
    py_ste_args = quantum_system.get_driving_pulses(*args)
    evolver = py_ste.get_unitary_evolver(H0, Hs)
    assert np.array_equal(value,
                          evolver.evolved_expectation_value(*py_ste_args, O))

def test_evolved_expectation_value_all():
    H0 = Z
    Hs = [X, Y]
    O = X+Y+Z
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    ctrl_amp_env = np.array([[1+2j, 4,     8    ],
                             [2,    7+ 5j, 9+10j],
                             [  3j,    6j,   11j],
                             [1,    1,     1    ]])
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    frequencies = np.array([0.5, 0, 2.5])
    number_channels = [1, 2]
    args = (ctrl_amp_env, initial_state, dt, frequencies, number_channels)
    value = quantum_system.evolved_expectation_value_all(*args, O)
    py_ste_args = quantum_system.get_driving_pulses(*args)
    evolver = py_ste.get_unitary_evolver(H0, Hs)
    assert np.array_equal(value,
                          evolver.evolved_expectation_value_all(*py_ste_args, O))

def test_ExpValCustom_initialisation():
    H0 = Z
    Hs = [X, Y]
    O = X+Y+Z
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    expval = ExpValCustom(quantum_system, initial_state, dt, O)
    
    assert expval.system == quantum_system
    assert np.array_equal(expval.initial_state, initial_state)
    assert expval.dt == dt
    assert np.array_equal(expval.observable, O)

def test_ExpValCustom_run():
    H0 = Z
    Hs = [X, Y]
    O = X+Y+Z
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    expval = ExpValCustom(quantum_system, initial_state, dt, O)
    ctrl_amp = np.array([[1+2j, 4,     8    ],
                         [2,    7+ 5j, 9+10j],
                         [  3j,    6j,   11j],
                         [1,    1,     1    ]])
    value = expval.run(tf.constant(ctrl_amp))
    value2 = quantum_system.evolver.evolved_expectation_value(ctrl_amp,
                                                              initial_state,
                                                              dt,
                                                              O)
    assert float(value) == float(value2.real)

def test_ExpValCustom_run_gradient():
    H0 = Z
    Hs = [X, Y]
    O = X+Y+Z
    hilbert_space = HilbertSpace([0, 1])
    quantum_system = QuantumSystem(H0, Hs, hilbert_space)
    dt = 0.1
    initial_state = np.array([1, 0], dtype=complex)
    expval = ExpValCustom(quantum_system, initial_state, dt, O)
    ctrl_amp = np.array([[1+2j, 4,     8    ],
                         [2,    7+ 5j, 9+10j],
                         [  3j,    6j,   11j],
                         [1,    1,     1    ]])
    x = tf.constant(ctrl_amp)
    with tf.GradientTape(persistent=False) as tape:
        tape.watch(x)
        value = expval.run(x)
    grad = tf.convert_to_tensor(tape.gradient(value, x)).numpy()
    grad2 = dt*quantum_system.evolver.switching_function(ctrl_amp,
                                                         initial_state,
                                                         dt,
                                                         O)[1]
    assert np.array_equal(grad, grad2)